package it.unibo.tw.dao.hsqldb;

import it.unibo.tw.dao.OrderDAO;
import it.unibo.tw.dao.OrderDTO;
import it.unibo.tw.dao.OrderItemDTO;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.List;
import java.util.logging.Logger;

/**
 * Implementazione di query jdbc su una versione hqldb della tabella entry
 * 
 * La scrittura dei metodi DAO con JDBC segue sempre e comunque questo schema:
 * 
	// --- 1. Dichiarazione della variabile per il risultato ---
	// --- 2. Controlli preliminari sui dati in ingresso ---
	// --- 3. Apertura della connessione ---
	// --- 4. Tentativo di accesso al db e impostazione del risultato ---
	// --- 5. Gestione di eventuali eccezioni ---
	// --- 6. Rilascio, SEMPRE E COMUNQUE, la connessione prima di restituire il controllo al chiamante
	// --- 7. Restituzione del risultato (eventualmente di fallimento)
 * 
 * Inoltre, all'interno di uno statement JDBC si segue lo schema generale
 * 
	// --- a. Crea (se senza parametri) o prepara (se con parametri) lo statement
	// --- b. Pulisci e imposta i parametri (se ve ne sono)
	// --- c. Esegui l'azione sul database ed estrai il risultato (se atteso)
	// --- d. Cicla sul risultato (se presente) pe accedere ai valori di ogni sua tupla
	// --- e. Rilascia la struttura dati del risultato
	// --- f. Rilascia la struttura dati dello statement
 * 
 * @author pisi79
 *
 */
public class HsqldbOrderDAO 
implements OrderDAO {

	Logger logger = Logger.getLogger( getClass().getCanonicalName() );

	// === Costanti letterali per non sbagliarsi a scrivere !!! ============================
	
	static final String TABLE = "orders";
	static final String JOIN_TABLE = "orders_items";

	// -------------------------------------------------------------------------------------

	static final String ID = "id";
	static final String BUYER = "buyer";
	static final String SHIPMENT = "shipment";
	static final String EMAIL = "email";
	
	static final String ORDERID = "orderId";
	static final String ITEMID = "itemId";
	static final String QUANTITY = "quantity";
	
	// == STATEMENT SQL ====================================================================

	// INSERT INTO table ( name,description, ...) VALUES ( ?,?, ... );
	static final String insert_order = 
		"INSERT " +
			"INTO " + TABLE + " ( " + 
				BUYER +", "+SHIPMENT+", "+EMAIL+" ) " +
			"VALUES (?,?,?) "
		;
	static final String insert_item = 
		"INSERT " +
			"INTO " + JOIN_TABLE + " ( " + 
				ORDERID+", "+ITEMID+", "+QUANTITY+" ) " +
			"VALUES (?,?,?) "
		;
	
	// SELECT * FROM table WHERE idcolumn = ?;
	static String read_order_by_id = 
		"SELECT O."+ID+", O."+BUYER+", O."+SHIPMENT+", O."+EMAIL+", OI."+ORDERID+", OI."+ITEMID+", OI."+QUANTITY+" FROM "+TABLE+" AS O LEFT JOIN "+JOIN_TABLE+" AS OI ON ( O."+ID+" = OI."+ORDERID+") WHERE O."+ID+" = ? ";

	// DELETE FROM table WHERE idcolumn = ?;
	static String delete = 
		"DELETE " +
			"FROM " + TABLE + " " +
			"WHERE " + ID + " = ? "
		;

	// SELECT * FROM table WHERE stringcolumn = ?;
	static String read_all = 
		"SELECT * " +
		"FROM " + TABLE + " ";
	
	static String read_by_buyer(String buyer) {
		return 
		read_all +
			"WHERE " + BUYER + " = '"+buyer+"' "
		;
	}
	
	// -------------------------------------------------------------------------------------

	// CALL IDENTITY();
	/**
	 * Restituisce l'identificativo dell'ultima tupla inserita. 
	 * TIPOCO ESEMPIO DI FUNZIONE >>>NON<<< PORTABILE: NON CROSS-DATABASE!!!
	 * Questa versione funziona su hibernate
	 */
	static String lastInsert = 
		"CALL " + 
			"IDENTITY() ;"
		;

	// -------------------------------------------------------------------------------------

	// CREATE entrytable ( idcolumn INT NOT NULL IDENTITY, ..., UNIQUE ( idcolumn ) );
	static String create = 
		"CREATE " +
			"TABLE " + TABLE +" ( " +
				ID + " BIGINT NOT NULL IDENTITY, " +
				BUYER + " VARCHAR(50), " +
				SHIPMENT + " LONGVARCHAR, " +
				EMAIL + " LONGVARCHAR " +
				", UNIQUE( " + ID + " ) " +
			")";
	static String create_jointable = 
		"CREATE " +
			"TABLE " + JOIN_TABLE +" ( " + ORDERID + " BIGINT NOT NULL, " + ITEMID + " BIGINT NOT NULL, "+ QUANTITY +" INT )";
	static String drop = 
		"DROP " +
			"TABLE " + TABLE + " "
		;
	static String drop_jointable = 
		"DROP " +
			"TABLE " + JOIN_TABLE + ""
		;
	
	// === METODI DAO =========================================================================
	
	/**
	 * C
	 */
	public Long create(OrderDTO t) {
		// === ORDINE ========================================
		
		// --- 1. Dichiarazione della variabile per il risultato ---
		Long result_orderID = new Long(-1);
		// --- 2. Controlli preliminari sui dati in ingresso ---
		if ( t == null )  {
			logger.warning("create(): failed to insert a null entry");
			return result_orderID;
		}
		// --- 3. Apertura della connessione ---
		Connection conn = HsqldbDAOFactory.createConnection();
		// --- 4. Tentativo di accesso al db e impostazione del risultato ---
		try {
			// --- a. Crea (se senza parametri) o prepara (se con parametri) lo statement
			PreparedStatement prep_stmt = conn.prepareStatement(HsqldbOrderDAO.insert_order);
			// --- b. Pulisci e imposta i parametri (se ve ne sono)
			prep_stmt.clearParameters();
			prep_stmt.setString(1, t.getBuyer());
			prep_stmt.setString(2, t.getShipment());
			prep_stmt.setString(3, t.getEmail());
			// --- c. Esegui l'azione sul database ed estrai il risultato (se atteso)
			prep_stmt.executeUpdate();
			// --- d. Cicla sul risultato (se presente) pe accedere ai valori di ogni sua tupla
			// n.d.
			// --- e. Rilascia la struttura dati del risultato
			// n.d.
			// --- f. Rilascia la struttura dati dello statement
			prep_stmt.close();
		}
		// --- 5. Gestione di eventuali eccezioni ---
		catch (Exception e) {
			logger.warning("create(): failed to insert entry: " + e.getMessage());
			e.printStackTrace();
			result_orderID = new Long(-2);
		}

		// Nel caso della creazione di una nuova tupla eseguo un secondo accesso per sapere che id le è stato assegnato
		// Chiaramente è inutile farlo se già il primo accesso ha prodotto errori
		// --- 6./7. Chiusura della connessione in caso di errori e restituizione prematura di un risultato di fallimento
		if ( result_orderID == -2 ) {
			HsqldbDAOFactory.closeConnection(conn);
			return result_orderID;
		}

		// === ID ========================================

		// --- 1. Dichiarazione della variabile per il risultato ---
		// riutilizziamo quella di prima

		// --- 2. Controlli preliminari sui dati in ingresso ---
		// già fatti

		// --- 3. Apertura della connessione ---
		// ce n'è una già aperta, se siamo qui

		// --- 4. Tentativo di accesso al db e impostazione del risultato ---
		try {
			Statement stmt = conn.createStatement();
			ResultSet rs = stmt.executeQuery(HsqldbOrderDAO.lastInsert);
			if ( rs.next() ) {
				result_orderID = rs.getLong(1);
			}		
			rs.close();
			stmt.close();
		}
		// --- 5. Gestione di eventuali eccezioni ---
		catch (Exception e) {
			logger.warning("create(): failed to retrieve id of last inserted entry: "+e.getMessage());
			e.printStackTrace();
		}
		// --- 6. Rilascio, SEMPRE E COMUNQUE, la connessione prima di restituire il controllo al chiamante
		// n.d.
		// --- 7. Restituzione del risultato (eventualmente di fallimento)
		// n.d.

		// === ITEMS ========================================

		// --- 1. Dichiarazione della variabile per il risultato ---
		// n.d.
		// --- 2. Controlli preliminari sui dati in ingresso ---
		if ( t.getItems() == null )  {
			return result_orderID;
		}
		// --- 3. Apertura della connessione ---
		conn = HsqldbDAOFactory.createConnection();
		// --- 4. Tentativo di accesso al db e impostazione del risultato ---
		for ( OrderItemDTO oi : t.getItems() ) {
			try {
				// --- a. Crea (se senza parametri) o prepara (se con parametri) lo statement
				PreparedStatement prep_stmt = conn.prepareStatement(HsqldbOrderDAO.insert_item);
				// --- b. Pulisci e imposta i parametri (se ve ne sono)
				prep_stmt.clearParameters();
				prep_stmt.setLong(1, result_orderID);
				prep_stmt.setLong(2, oi.getItem().getId());
				prep_stmt.setInt(3 , oi.getQuantity());
				// --- c. Esegui l'azione sul database ed estrai il risultato (se atteso)
				prep_stmt.executeUpdate();
				// --- d. Cicla sul risultato (se presente) pe accedere ai valori di ogni sua tupla
				// n.d.
				// --- e. Rilascia la struttura dati del risultato
				// n.d.
				// --- f. Rilascia la struttura dati dello statement
				prep_stmt.close();
			}
			// --- 5. Gestione di eventuali eccezioni ---
			catch (Exception e) {
				logger.warning("create(): failed to insert entry: " + e.getMessage());
				e.printStackTrace();
				// in a real webapp you should rollback your transaction, here! Not really just behave as follows...
				this.delete(result_orderID);
				result_orderID = new Long(-2);
				break;
			}
		}

		// Nel caso della creazione di una nuova tupla eseguo un secondo accesso per sapere che id le è stato assegnato
		// Chiaramente è inutile farlo se già il primo accesso ha prodotto errori
		// --- 6. Rilascio, SEMPRE E COMUNQUE, la connessione prima di restituire il controllo al chiamante
		HsqldbDAOFactory.closeConnection(conn);
		// --- 7. Restituzione del risultato (eventualmente di fallimento)
		return result_orderID;

	}

	// -------------------------------------------------------------------------------------

	/**
	 * R
	 */
	public OrderDTO read(Long id) {
		// --- 1. Dichiarazione della variabile per il risultato ---
		OrderDTO result = null;
		// --- 2. Controlli preliminari sui dati in ingresso ---
		if ( id < 0 )  {
			logger.warning("read(): cannot read an entry with a negative id");
			return result;
		}
		// --- 3. Apertura della connessione ---
		Connection conn = HsqldbDAOFactory.createConnection();
		// --- 4. Tentativo di accesso al db e impostazione del risultato ---
		try {
			// --- a. Crea (se senza parametri) o prepara (se con parametri) lo statement
			PreparedStatement prep_stmt = conn.prepareStatement(HsqldbOrderDAO.read_order_by_id);
			// --- b. Pulisci e imposta i parametri (se ve ne sono)
			prep_stmt.clearParameters();
			prep_stmt.setLong(1, id);
			// --- c. Esegui l'azione sul database ed estrai il risultato (se atteso)
			ResultSet rs = prep_stmt.executeQuery();
			// --- d. Cicla sul risultato (se presente) pe accedere ai valori di ogni sua tupla
			// JDBC ci obbliga gestire a mano il join e ricreare gli oggetti di conseguenza!
			ResultSet set = prep_stmt.executeQuery();
			while (set.next()) {
				if ( result == null )
					result = new OrderDTO();
				if (result.getId() == null) {
					result.setId(set.getLong(ID));
					result.setBuyer(set.getString(BUYER));
					result.setShipment(set.getString(SHIPMENT));
					result.setEmail(set.getString(EMAIL));
					if ( result.getItems() == null )
						result.setItems( new ArrayList<OrderItemDTO>() );
				}
				if ( set.getLong(ORDERID) >= 0 ) {
					OrderItemDTO oi = new OrderItemDTO();
					oi.setItemDTO( new HsqldbItemDAO().read( set.getLong(ITEMID)));
					oi.setQuantity( set.getInt(QUANTITY));
					result.getItems().add(oi);
				}
			}
			set.close();
			prep_stmt.close();			
			
			// --- e. Rilascia la struttura dati del risultato
			rs.close();
			// --- f. Rilascia la struttura dati dello statement
			prep_stmt.close();
		}
		// --- 5. Gestione di eventuali eccezioni ---
		catch (Exception e) {
			logger.warning("read(): failed to retrieve entry with id = " + id+": "+e.getMessage());
			e.printStackTrace();
		}
		// --- 6. Rilascio, SEMPRE E COMUNQUE, la connessione prima di restituire il controllo al chiamante
		finally {
			HsqldbDAOFactory.closeConnection(conn);
		}
		// --- 7. Restituzione del risultato (eventualmente di fallimento)
		return result;
	}

	// -------------------------------------------------------------------------------------

	/**
	 * D
	 */
	public boolean delete(Long id) {
		// --- 1. Dichiarazione della variabile per il risultato ---
		boolean result = false;
		// --- 2. Controlli preliminari sui dati in ingresso ---
		if ( id == null || id < 0 )  {
			logger.warning("delete(): cannot delete an entry with an invalid id ");
			return result;
		}
		// --- 3. Apertura della connessione ---
		Connection conn = HsqldbDAOFactory.createConnection();
		// --- 4. Tentativo di accesso al db e impostazione del risultato ---
		try {
			// --- a. Crea (se senza parametri) o prepara (se con parametri) lo statement
			PreparedStatement prep_stmt = conn.prepareStatement(HsqldbOrderDAO.delete);
			// --- b. Pulisci e imposta i parametri (se ve ne sono)
			prep_stmt.clearParameters();
			prep_stmt.setLong(1, id);
			// --- c. Esegui l'azione sul database ed estrai il risultato (se atteso)
			prep_stmt.executeUpdate();
			// --- d. Cicla sul risultato (se presente) pe accedere ai valori di ogni sua tupla
			// n.d. Qui devo solo dire al chiamante che è andato tutto liscio
			result = true;
			// --- e. Rilascia la struttura dati del risultato
			// n.d.
			// --- f. Rilascia la struttura dati dello statement
			prep_stmt.close();
		}
		// --- 5. Gestione di eventuali eccezioni ---
		catch (Exception e) {
			logger.warning("delete(): failed to delete entry with id = " + id+": "+e.getMessage());
			e.printStackTrace();
		}
		// --- 6. Rilascio, SEMPRE E COMUNQUE, la connessione prima di restituire il controllo al chiamante
		finally {
			HsqldbDAOFactory.closeConnection(conn);
		}
		// --- 7. Restituzione del risultato (eventualmente di fallimento)
		return result;
	}

	// -------------------------------------------------------------------------------------

	/**
	 * Find by name
	 */
	public List<OrderDTO> find(String buyer) {
		// --- 1. Dichiarazione della variabile per il risultato ---
		List<OrderDTO> result = new ArrayList<OrderDTO>();
		// --- 2. Controlli preliminari sui dati in ingresso ---
		boolean any = false;
		if ( buyer == null || buyer.equals("") )
			any = true;
		// --- 3. Apertura della connessione ---
		Connection conn = HsqldbDAOFactory.createConnection();
		// --- 4. Tentativo di accesso al db e impostazione del risultato ---
		try {
			// --- a. Crea (se senza parametri) o prepara (se con parametri) lo statement
			Statement stmt = conn.createStatement();
			// --- b. Pulisci e imposta i parametri (se ve ne sono)
			// n.d.
			// --- c. Esegui l'azione sul database ed estrai il risultato (se atteso)
			ResultSet rs = stmt.executeQuery( any ? HsqldbOrderDAO.read_all : HsqldbOrderDAO.read_by_buyer(buyer) );
			// --- d. Cicla sul risultato (se presente) pe accedere ai valori di ogni sua tupla
			while ( rs.next() ) {
				OrderDTO entry = new OrderDTO();
				entry.setId(rs.getLong(ID));
				entry.setBuyer(rs.getString(BUYER));
				entry.setShipment(rs.getString(SHIPMENT));
				entry.setEmail(rs.getString(EMAIL));
				result.add( entry );
			}
			// --- e. Rilascia la struttura dati del risultato
			rs.close();
			// --- f. Rilascia la struttura dati dello statement
			stmt.close();
		}
		// --- 5. Gestione di eventuali eccezioni ---
		catch (Exception e) {
			logger.warning("find(): failed to retrieve enttry with buyer = '" + buyer +"': "+e.getMessage());
			e.printStackTrace();
		}
		// --- 6. Rilascio, SEMPRE E COMUNQUE, la connessione prima di restituire il controllo al chiamante
		finally {
			HsqldbDAOFactory.closeConnection(conn);
		}
		// --- 7. Restituzione del risultato (eventualmente di fallimento)
		return result;
	}

	// -------------------------------------------------------------------------------------
	// -------------------------------------------------------------------------------------
	// -------------------------------------------------------------------------------------

	/**
	 * Creazione della table
	 */
	public boolean createTable() {
		// --- 1. Dichiarazione della variabile per il risultato ---
		boolean result = false;
		// --- 2. Controlli preliminari sui dati in ingresso ---
		// n.d.
		// --- 3. Apertura della connessione ---
		Connection conn = HsqldbDAOFactory.createConnection();
		// --- 4. Tentativo di accesso al db e impostazione del risultato ---
		try {
			// --- a. Crea (se senza parametri) o prepara (se con parametri) lo statement
			Statement stmt = conn.createStatement();
			// --- b. Pulisci e imposta i parametri (se ve ne sono)
			// n.d.
			// --- c. Esegui l'azione sul database ed estrai il risultato (se atteso)
			stmt.execute(create);
			// --- d. Cicla sul risultato (se presente) pe accedere ai valori di ogni sua tupla
			// n.d. Qui devo solo dire al chiamante che è andato tutto liscio
			result = true;
			// --- e. Rilascia la struttura dati del risultato
			// n.d.
			// --- f. Rilascia la struttura dati dello statement
			stmt.close();
		}
		// --- 5. Gestione di eventuali eccezioni ---
		catch (Exception e) {
			logger.warning("createTable(): failed to create table '"+TABLE+"': "+e.getMessage());
		}
		try {
			// --- a. Crea (se senza parametri) o prepara (se con parametri) lo statement
			Statement stmt = conn.createStatement();
			// --- b. Pulisci e imposta i parametri (se ve ne sono)
			// n.d.
			// --- c. Esegui l'azione sul database ed estrai il risultato (se atteso)
			stmt.execute(create_jointable);
			// --- d. Cicla sul risultato (se presente) pe accedere ai valori di ogni sua tupla
			// n.d. Qui devo solo dire al chiamante che è andato tutto liscio
			result = result && true;
			// --- e. Rilascia la struttura dati del risultato
			// n.d.
			// --- f. Rilascia la struttura dati dello statement
			stmt.close();
		}
		// --- 5. Gestione di eventuali eccezioni ---
		catch (Exception e) {
			logger.warning("createTable(): failed to create table '"+JOIN_TABLE+"': "+e.getMessage());
		}
		// --- 6. Rilascio, SEMPRE E COMUNQUE, la connessione prima di restituire il controllo al chiamante
		finally {
			HsqldbDAOFactory.closeConnection(conn);
		}
		// --- 7. Restituzione del risultato (eventualmente di fallimento)
		return result;
	}

	// -------------------------------------------------------------------------------------

	/**
	 * Rimozione della table
	 */
	public boolean dropTable() {
		// --- 1. Dichiarazione della variabile per il risultato ---
		boolean result = false;
		// --- 2. Controlli preliminari sui dati in ingresso ---
		// n.d.
		// --- 3. Apertura della connessione ---
		Connection conn = HsqldbDAOFactory.createConnection();
		// --- 4. Tentativo di accesso al db e impostazione del risultato ---
		try {
			// --- a. Crea (se senza parametri) o prepara (se con parametri) lo statement
			Statement stmt = conn.createStatement();
			// --- b. Pulisci e imposta i parametri (se ve ne sono)
			// n.d.
			// --- c. Esegui l'azione sul database ed estrai il risultato (se atteso)
			stmt.execute(drop_jointable);
			// --- d. Cicla sul risultato (se presente) pe accedere ai valori di ogni sua tupla
			// n.d. Qui devo solo dire al chiamante che è andato tutto a posto.
			result = true;
			// --- e. Rilascia la struttura dati del risultato
			// n.d.
			// --- f. Rilascia la struttura dati dello statement
			stmt.close();
		}
		// --- 5. Gestione di eventuali eccezioni ---
		catch (Exception e) {
			logger.warning("dropTable(): failed to drop table '"+JOIN_TABLE+"': "+e.getMessage());
		}
		try {
			// --- a. Crea (se senza parametri) o prepara (se con parametri) lo statement
			Statement stmt = conn.createStatement();
			// --- b. Pulisci e imposta i parametri (se ve ne sono)
			// n.d.
			// --- c. Esegui l'azione sul database ed estrai il risultato (se atteso)
			stmt.execute(drop);
			// --- d. Cicla sul risultato (se presente) pe accedere ai valori di ogni sua tupla
			// n.d. Qui devo solo dire al chiamante che è andato tutto a posto.
			result = result && true;
			// --- e. Rilascia la struttura dati del risultato
			// n.d.
			// --- f. Rilascia la struttura dati dello statement
			stmt.close();
		}
		// --- 5. Gestione di eventuali eccezioni ---
		catch (Exception e) {
			logger.warning("dropTable(): failed to drop table '"+TABLE+"': "+e.getMessage());
		}
		// --- 6. Rilascio, SEMPRE E COMUNQUE, la connessione prima di restituire il controllo al chiamante
		finally {
			HsqldbDAOFactory.closeConnection(conn);
		}
		// --- 7. Restituzione del risultato (eventualmente di fallimento)
		return result;
	}

}